#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

#define RND_N 65537
#define RND_G 75
int seed = 1;

int rnd ()
{
  seed = RND_G * seed % RND_N;
  return seed;
}

void print_map (char **plane, int size) {
  for (int i = 0; i < size; i++) {
    for (int j = 0; j < size; j++)
      printf ("%c", (plane[i][j] == 1) ? 'O' : '.');
    printf ("\n");
  }
  printf ("\n");
}

void print_ppm (char **plane, int size, char *file) {
  FILE *f = fopen (file, "w");
  int sq = 8;
  fprintf (f, "P3\n");
  fprintf (f, "%d %d\n",
           size * sq + (size - 1) + 2, size * sq + (size - 1) + 2);
  fprintf (f, "255\n");
  for (int i = 0; i < size * sq + (size - 1) + 2; i++) {
    for (int j = 0; j < size * sq + (size - 1) + 2; j++) {
      int ii = i / (sq + 1);
      int jj = j / (sq + 1);
      int c;
      if (i % (sq + 1) == 0) c = 200; else
      if (j % (sq + 1) == 0) c = 200; else
      if (plane[ii][jj] == 0) c = 255; else c = 0;
      fprintf (f, "%d %d %d ", c, c, c);
    }
    fprintf (f, "\n");
  }
  fclose (f);
}

int neighbors (char **plane, int size, int i, int j) {
  int neighs = 0;
  for (int ii = -1; ii <= +1; ii++)
    if ((i + ii >= 0) && (i + ii < size))
      for (int jj = -1; jj <= +1; jj++)
        if ((j + jj >= 0) && (j + jj < size))
          if ((ii != 0) || (jj != 0))
            neighs += ((int)plane[i + ii][j + jj]);
  return neighs;
}

int main (int argc, char *argv[]) {
  int size = 10;
  int gens = 1;
  if (argc < 1) return 1;
  if ((argc >= 2) && (sscanf (argv[1], "%d", &size) != 1)) return 1;
  if ((argc >= 3) && (sscanf (argv[2], "%d", &seed) != 1)) return 1;
  if ((argc >= 4) && (sscanf (argv[3], "%d", &gens) != 1)) return 1;
  if (argc >= 5) return 1;

  char **plane = (char**)malloc (size * sizeof (char*));
  char **aux_plane = (char**)malloc (size * sizeof (char*));
  for (int i = 0; i < size; i++) {
    plane[i] = (char*)malloc (size * sizeof (char));
    aux_plane[i] = (char*)malloc (size * sizeof (char));
  }

  for (int i = 0; i < size; i++)
    for (int j = 0; j < size; j++) {
      plane[i][j] = rnd () %2;
      aux_plane[i][j] = 0;
    }
  //print_ppm (plane, size, "life-fstgen.ppm");
#if DEBUG==1
  print_map (plane, size);
#endif

  double t = omp_get_wtime ();
  #pragma omp parallel firstprivate(size,gens,plane,aux_plane)
  {
    while (gens > 0) {
      #pragma omp for collapse(2)
        for (int i = 0; i < size; i++)
          for (int j = 0; j < size; j++) {
            int neighs = neighbors (plane, size, i, j);
            switch (plane[i][j]) {
              case 0: aux_plane[i][j] = (neighs == 3);
                      break;
              case 1: aux_plane[i][j] = (neighs == 2) || (neighs == 3);
                      break;
            }
          }
      gens--;
      char **tmp_plane = aux_plane; aux_plane = plane; plane = tmp_plane;
      #pragma omp barrier
    }
  }
  printf ("TIME: %lf\n", omp_get_wtime () - t);

  //print_ppm (plane, size, "life-lstgen.ppm");
#if DEBUG==1
  print_map (plane, size);
#endif

  return 0;
}
